#!/usr/bin/perl

#Kevin Keating
#MB&B 752b: Genomics and Bioinformatics
#Spring 2005
#Final project

#trainHexFreq
#Input:
#       A file containing hexamer probabilities (generated by trainHexFreq.pl
#       A file containing the DNA sequence to be predicted
#           (This file may be in any format readable by BioPerl)
#Output:
#       A list of predicted coding regions
#Usage:
#       predict.pl hex-prob.file sequence.file
#
#
#This program predicts exons in human DNA using a two-state hidden Markov
#model based on hexamer frequencies.  A more detailed description of the
#algorithm may be found in readme.txt


use strict;                   #force variable declarations
use English '-no-match-vars'; #use long variable names, but don't use $PREMATCH and $POSTMATCH (since these slow down Perl and are not needed)
use Bio::SeqIO;               #BioPerl module for reading sequence data


#define integer values for the coding and non-coding states
sub noncoding () { 0 }
sub coding    () { 1 }

#define values for the HMM parameters

#start probabilities - the pi matrix
my $startLProb = [];
$startLProb->[noncoding] = log(.984);
$startLProb->[coding]    = log(.016);

#transition probabilities - the a matrix
my $transLProb = [];
$transLProb->[noncoding]->[noncoding] = log(.9994); #average intron (or other non-coding) region is roughly 5000 bp = 1667 "codons" long
$transLProb->[noncoding]->[coding]    = log(.0006); #1/1667 = .0006
$transLProb->[coding]   ->[noncoding] = log(.0171); #average exon is roughly 175 bp = 58 codons long
$transLProb->[coding]   ->[coding]    = log(.9829); #1/58 = .0171

#read in the emission probabilities (b matrix) from the Data::Dumper file generated by trainHexFreq.pl
my $emitLProb = [];
{
    my ($nonHex, $codHex);
    open(INP, $ARGV[0]);
    my $hexProbs = join("", (<INP>));
    close(INP);
    eval($hexProbs);
    
    $emitLProb->[noncoding] = $nonHex;
    $emitLProb->[coding]    = $codHex;
}


#read the sequence file containing the genes to predict
my $inputFile = Bio::SeqIO->new(-file => $ARGV[1]);
my $ob = $inputFile->next_seq;   #read in the observed DNA sequence

#print out the header for the output table
print "Predicted Exons:\n";
print "Start bp\t End bp\n";


#examine each forward reading frame
foreach my $curFrame (0..2) {
    #run the Viterbi algorithm on the observed DNA sequence
    my $pathLProbs = []; #the log probability of having followed a given path (the delta or phi array)
    my $path      = []; #the path followed for $pathProbs (the psi array)
        #accessed as ->[t]->[i] for hexamer t and state (coding or noncoding) i
    
    #initialize the $pathProbs array
    $pathLProbs->[1]->[noncoding] = $startLProb->[noncoding] + $emitLProb->[noncoding]->{getHex(1,$ob, $curFrame)};
    $pathLProbs->[1]->[coding]    = $startLProb->[coding]    + $emitLProb->[coding]   ->{getHex(1,$ob, $curFrame)};
    
    #iteratively fill in the $pathProbs array
    my $curHexNum = 2; #the hexamer currently being examined
    while (($curHexNum+1)*3 + $curFrame < $ob->length()) {
        my $probNN = $pathLProbs->[$curHexNum-1]->[noncoding] + $transLProb->[noncoding]->[noncoding]; #the chance of going from noncoding to noncoding
        my $probCN = $pathLProbs->[$curHexNum-1]->[coding]    + $transLProb->[coding]   ->[noncoding]; #the chance of going from coding    to noncoding
        
        my $probNC = $pathLProbs->[$curHexNum-1]->[noncoding] + $transLProb->[noncoding]->[coding];    #the chance of going from noncoding to coding
        my $probCC = $pathLProbs->[$curHexNum-1]->[coding]    + $transLProb->[coding]   ->[coding];    #the chance of going from coding    to coding
        
        #determine the most likely path to end up in the noncoding state
        if ($probNN > $probCN) {
            #if the noncoding path was more likely
            $pathLProbs->[$curHexNum]->[noncoding] = $probNN + $emitLProb->[noncoding]->{getHex($curHexNum, $ob, $curFrame)};
            $path      ->[$curHexNum]->[noncoding] = noncoding;
        } else {
            #if the coding path was more likely
            $pathLProbs->[$curHexNum]->[noncoding] = $probCN + $emitLProb->[noncoding]->{getHex($curHexNum, $ob, $curFrame)};
            $path      ->[$curHexNum]->[noncoding] = coding;
        }
        
        #determine the most likely path to end up in the coding state
        if ($probNC > $probCC) {
            #if the noncoding path was more likely
            $pathLProbs->[$curHexNum]->[coding] = $probNC + $emitLProb->[coding]->{getHex($curHexNum, $ob)};
            $path      ->[$curHexNum]->[coding] = noncoding;
        } else {
            #if the coding path was more likely
            $pathLProbs->[$curHexNum]->[coding] = $probCC + $emitLProb->[coding]->{getHex($curHexNum, $ob)};
            $path      ->[$curHexNum]->[coding] = coding;
        }
        
        ++$curHexNum; #move to the next hexamer
    }
    
    #determine what the best path was
    my $bestPath = [];
    
    #find the most likely end state
    my$lastHexNum = --$curHexNum; #the number of the last hexamer
    if ($pathLProbs->[$curHexNum]->[noncoding] > $pathLProbs->[$curHexNum]->[coding]) {
        #if the noncoding state was more likely
        $bestPath->[$curHexNum] = noncoding;
    } else {
        #if the coding state was more likely
        $bestPath->[$curHexNum] = coding;
    }
    
    #follow the path back
    while ($curHexNum > 1) {
        --$curHexNum;
        $bestPath->[$curHexNum] = $path->[$curHexNum + 1]->[$bestPath->[$curHexNum+1]];
    }
    
    #print out the predicted coding regions
    if ($bestPath->[1] == coding) {
        print 1 + $curFrame . "\t";
    }
    for $curHexNum (2..$lastHexNum) {
        if      ($bestPath->[$curHexNum] == coding    and $bestPath->[$curHexNum-1] == noncoding) {
            print $curHexNum * 3 - 2 + $curFrame . "\t";
        } elsif ($bestPath->[$curHexNum] == noncoding and $bestPath->[$curHexNum-1] == coding   ) {
            print $curHexNum * 3 - 2 + $curFrame . "\n";
        }
    }
    if ($bestPath->[$lastHexNum] == coding) {
        print $lastHexNum * 3 - 2 . + $curFrame . "\n";
    }
}


sub getHex {
    #return the DNA sequence for the requested hexamer
    my $hexNum = shift;   #the number of the hexamer to get
    my $ob     = shift;   #the observed DNA sequence
    my $curFrame = shift; #the current readingframe
    
    return $ob->subseq($hexNum*3 - 2 + $curFrame, $hexNum*3 + 3 + $curFrame);
}