#!/usr/bin/perl

#Kevin Keating
#MB&B 752b: Genomics and Bioinformatics
#Spring 2005
#Final project

#trainHexFreq
#Input:
#       A file containing a sample of non-coding human DNA
#       A file containing a sample of coding human DNA
#       (These files may be in any format readable by BioPerl)
#Output:
#       A Data::Dumper file container hexamer usage log probabilities for
#       coding and non-coding regions, based on the input files
#Usage:
#       trainHexFreq.pl non-coding.file coding.file output.file
#
#This program calculates hexamer usage in coding vs. non-coding DNA.  A more
#detailed description of the algorithm may be found in readme.txt

use strict;                   #force variable declarations
use English '-no-match-vars'; #use long variable names, but don't use $PREMATCH and $POSTMATCH (since these slow down Perl and are not needed)
use Data::Dumper;             #used to output hexamer usage results to a file
use Bio::SeqIO;               #BioPerl module for reading sequence data


#examine the non-coding region
my $noncoding = Bio::SeqIO->new(-file => $ARGV[0]); #open the non-coding file
my $nonHex = {};    #the frequencies of any given hexamer in a non-coding region

#count hexamers in each sequence of the noncoding file
print "Computing hexamers in noncoding sequence...\n";
#examine each sequence
while (my $curSeq = $noncoding->next_seq) {
    #examine each hexamer in the current sequence
    foreach my $curStart (1..($curSeq->length()-5)) {
        print "\t$curStart\n" unless ($curStart % 1000); #print out status reports every 1000 bp
        my $hexamer = $curSeq->subseq($curStart, $curStart + 5);
        next if ($hexamer =~ /[NnXx]/); #skip hexamers that have unknown bases
        ++$nonHex->{uc($hexamer)};
    }
}

#count the total number of noncoding hexamers
my $totNonHex; #the total number of noncoding hexamers
foreach my $curCount (values(%{$nonHex})) {
    $totNonHex += $curCount;
}

#convert the counts of hexamers to log probabilities
while (my ($curHex, $curCount) = each(%{$nonHex})) {
    $nonHex->{$curHex} = log($curCount / $totNonHex);
}


#examine the coding region
my $coding = Bio::SeqIO->new(-file => $ARGV[1]); #open the coding file
my $codHex = {};    #the frequencies of any given hexamer in a coding region

#count hexamers in each sequence of the noncoding file
print "Computing hexamers in coding sequence...\n";
my $count = 0;  #used to print out status reports
while (my $curSeq = $coding->next_seq) {
    print "\t$count\n" unless ($count++ % 100); #print out status reports every 100 sequences

    my $curStart = 1; #the starting base of the current hexamer
    while ($curStart < ($curSeq->length()-5)) { #only read in-frame hexamers
        my $hexamer = $curSeq->subseq($curStart, $curStart + 5);
        $curStart += 3; #move to the next in-frame read
        next if ($hexamer =~ /[NnXx]/); #skip hexamers that have unknown bases
        ++$codHex->{uc($hexamer)};
    }
}

#count the total number of coding hexamers
my $totCodHex; #the total number of coding hexamers
foreach my $curCount (values(%{$codHex})) {
    $totCodHex += $curCount;
}

#convert the counts of hexamers to log probabilities
while (my ($curHex, $curCount) = each(%{$codHex})) {
    $codHex->{$curHex} = log($curCount / $totCodHex);
}


#store the log probabilities in an output file
open(OUT, ">", $ARGV[2]) or die "Could not open file for output\n";
print OUT Data::Dumper->Dump([$nonHex, $codHex], ["nonHex", "codHex"]);
close(OUT);